from typing import Any, Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.model_selection import train_test_split
from statsmodels.tsa.seasonal import seasonal_decompose

# Specifying custom types
TrainValidTest = Tuple[
    pd.DataFrame, pd.Series, pd.DataFrame, pd.Series, pd.DataFrame, pd.Series
]
TrainTest = Tuple[pd.DataFrame, pd.Series, pd.DataFrame, pd.Series]


class FeaturesExtractor(BaseEstimator, TransformerMixin):
    """Implements feature engineering for time series DataFrame."""

    def __init__(
        self,
        max_lag: int = 1,
        rolling_mean_order: int = 1,
    ) -> None:
        """Constructor for FeaturesExtractor class."""
        self.max_lag = max_lag
        self.rolling_mean_order = rolling_mean_order

    def fit(self, x: pd.DataFrame, y=None) -> None:
        """Returns the object itself."""
        return self

    def transform(self, x: pd.DataFrame, y=None) -> pd.DataFrame:
        """
        Extracts features from DateTimeIndex in accordance
        with arguments passed to `__init__`.
        """
        X = x.copy()
        # Adding time variables
        X["quarter"] = X.index.quarter
        X["month"] = X.index.month
        X["day"] = X.index.day
        X["hour"] = X.index.hour
        # Adding lags
        for lag in range(1, self.max_lag + 1):
            X["lag_{}".format(lag)] = X["num_orders"].shift(lag)
        # Adding moving average
        X["rolling_mean_{}".format(self.rolling_mean_order)] = (
            X["num_orders"].shift().rolling(self.rolling_mean_order).mean()
        )

        return X


def prepare_ts(
    data: pd.DataFrame,
    target_name: str,
    train_share: float,
    test_share: float,
    valid_share: Optional[float] = None,
) -> Union[TrainValidTest, TrainTest]:
    """Conducts time series data split into sets.

    Depending on the value passed to `valid_share`,
    splits the data either into training, validation
    and test sets or into training and test sets.

    Args:
        data (pd.DataFrame): DataFrame which needs to be
            split into sets.
        target_name (str): Name of the target variable.
        train_share (float): Share of the training set.
        test_share (float): Share of the test set.
        valid_share (Optional[float], optional): Share of
            the validation set. Defaults to None.

    Returns:
        Union[TrainValidTest, TrainTest]: Tuple of training,
        validation and test sets or Tuple of training and
        test sets.

    Raises:
        ValueError: Exception raised in case proportions
        of sets are inconsistent.
    """
    # Data split into train/validation/test
    if valid_share is not None:
        # Checking consistency of set shares
        if np.sum([train_share, valid_share, test_share]) != 1.0:
            raise ValueError("Incorrect sets proportions specified")

        split_1 = 1 - train_share
        split_2 = test_share / np.sum([valid_share, test_share])

        training_set, valid_test_set = train_test_split(
            data, shuffle=False, test_size=split_1
        )
        valid_set, test_set = train_test_split(
            valid_test_set, shuffle=False, test_size=split_2
        )

        # Dropping nans due to lags
        training_set = training_set.dropna()

        features_train = training_set.drop([target_name], axis=1)
        target_train = training_set[target_name]

        features_valid = valid_set.drop([target_name], axis=1)
        target_valid = valid_set[target_name]

        features_test = test_set.drop([target_name], axis=1)
        target_test = test_set[target_name]

        return (
            features_train,
            target_train,
            features_valid,
            target_valid,
            features_test,
            target_test,
        )

    # Data split into training/test
    if np.sum([train_share, test_share]) != 1.0:
        raise ValueError("Incorrect sets proportions specified")

    # Data split
    training_set, test_set = train_test_split(
        data, shuffle=False, test_size=test_share
    )

    # Dropping nans due to lags
    training_set = training_set.dropna()

    # Separating features from target
    features_train = training_set.drop([target_name], axis=1)
    target_train = training_set[target_name]

    features_test = test_set.drop([target_name], axis=1)
    target_test = test_set[target_name]

    return features_train, target_train, features_test, target_test


def plot_time_series(
    data: pd.DataFrame,
    col: str,
    period_start: str,
    period_end: str,
    kind: Optional[str] = None,
) -> Any:
    """Plots time series.

    When specifying kind="decomposed", conducts seasonal decomposition
    of the data and outputs graphs of trend, seasonal and residual
    components.

    Args:
        data (pd.DataFrame): DataFrame with information
            about taxi orders.
        col (str): DataFrame column containing
            data about taxi orders.
        period_start (str): Time period start.
        period_end (str): Time period end.
        kind (Optional[str], optional): Boolean indicator
            of performing time-series decomposition.
    """
    # Plotting decomposed time series
    if kind == "decomposed":
        decomposed = seasonal_decompose(data)

        plt.figure(figsize=(8, 10))

        # Trend component
        plt.subplot(311)
        trend_plot = sns.lineplot(
            data=decomposed.trend[period_start:period_end], ax=plt.gca()
        )
        trend_plot.set(title="Trend", xlabel="Time period", ylabel="Orders")
        plt.xticks(rotation=45)

        # Seasonal component
        plt.subplot(312)
        seasonal_plot = sns.lineplot(
            data=decomposed.seasonal[period_start:period_end], ax=plt.gca()
        )
        seasonal_plot.set(
            title="Seasonality", xlabel="Time period", ylabel="Orders"
        )
        plt.xticks(rotation=45)

        # Residual component
        plt.subplot(313)
        residual_plot = sns.lineplot(
            data=decomposed.resid[period_start:period_end], ax=plt.gca()
        )
        residual_plot.set(
            title="Residual", xlabel="Time period", ylabel="Orders"
        )
        plt.xticks(rotation=45)

        plt.tight_layout()
        plt.show()

        return

    # Plotting time series
    full_data_plot = sns.lineplot(
        data=data[period_start:period_end],
        y=col,
        x=data[period_start:period_end].index,
        ax=plt.gca(),
    )
    full_data_plot.set(
        title="Taxi orders number", xlabel="Time period", ylabel="Orders"
    )
    plt.xticks(rotation=45)
    plt.tight_layout()
    plt.show()



from typing import Optional, Tuple, Union

import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split

# Specifying custom types
TrainValidTest = Tuple[
    pd.DataFrame, pd.Series, pd.DataFrame, pd.Series, pd.DataFrame, pd.Series
]
TrainTest = Tuple[pd.DataFrame, pd.Series, pd.DataFrame, pd.Series]


def identify_missing_values(data: pd.DataFrame) -> Optional[pd.DataFrame]:
    """Performs missing values computation.

    Computes a number and share of missing values
    in DataFrame columns which have NaN-values present
    and displays data types of such columns.

    Args:
        data (pd.DataFrame): DataFrame which needs to
            be checked for missing values.

    Returns:
        Optional[pd.DataFrame]: DataFrame with column names, number of
        missing values and shares of NaN-values in such columns.
        In case there are not missing values present, an according
        message is shown and nothing is returned. Additionally, columns
        data type is shown.
    """
    # Verifying missing values
    miss_vals_num = data.isnull().sum()[data.isnull().sum() > 0]
    if miss_vals_num.empty:
        print("Missing values are not found.")
        return
    # Creating a table with numbers of missing values
    cols = {"missing_count": miss_vals_num.values}
    nans_df = pd.DataFrame(data=cols, index=miss_vals_num.index).sort_values(
        by="missing_count", ascending=False
    )
    # Adding shares of missing values
    nans_df["missing_fraction"] = nans_df["missing_count"] / data.shape[0]
    nans_df["missing_fraction"] = nans_df["missing_fraction"].round(4)
    # Adding data types
    nans_df["dtype"] = data[nans_df.index].dtypes
    nans_df = nans_df[["dtype", "missing_count", "missing_fraction"]]

    return nans_df


def preprocess_duplicates(data: pd.DataFrame) -> Optional[pd.DataFrame]:
    """Computes and deletes all duplicates in the DataFrame.

    Args:
        data (pd.DataFrame): DataFrame which needs to
            be checked for identical rows.
    """
    num_duplicates = data.duplicated().sum()
    if num_duplicates != 0:
        data.drop_duplicates(inplace=True)
        print(f"{num_duplicates:,} duplicates found and deleted.")
    else:
        print("No duplicates found.")


def prepare_sets(
    data: pd.DataFrame,
    target_name: str,
    train_share: float,
    test_share: float,
    valid_share: Optional[float] = None,
    random_state: Optional[int] = None,
) -> Union[TrainValidTest, TrainTest]:
    """
    Conducts data split into sets.

    Depending on the value passed to `valid_share`,
    splits the data either into training, validation
    and test sets or into training and test sets.

    Args:
        data (pd.DataFrame): DataFrame which needs to be split into sets.
        target_name (str): Name of the target variable.
        train_share (float): Share of the training set.
        test_share (float): Share of the test set.
        valid_share (Optional[float], optional): Share of the validation set. Defaults to None.
        random_state (Optional[int], optional): Random seed. Defaults to None.

    Returns:
        Union[TrainValidTest, TrainTest]: Tuple of training,
        validation and test sets or Tuple of training and
        test sets.

    Raises:
        ValueError: Exception raised in case proportions
        of sets are inconsistent.
    """
    # Separating features from target
    features = data.drop([target_name], axis=1)
    target = data[target_name]

    # Data split into sets with a validation set
    if valid_share is not None:
        # Checking consistency of set shares
        if np.sum([train_share, valid_share, test_share]) != 1.0:
            raise ValueError("Incorrect sets proportions specified")

        # Determining cutoffs
        split_1 = 1 - train_share
        split_2 = test_share / np.sum([valid_share, test_share])

        # Splitting into training set and a combination of validation/test sets
        (
            features_train,
            features_valid_test,
            target_train,
            target_valid_test,
        ) = train_test_split(
            features, target, test_size=split_1, random_state=random_state
        )

        # Separating validation set from test set
        (
            features_valid,
            features_test,
            target_valid,
            target_test,
        ) = train_test_split(
            features_valid_test,
            target_valid_test,
            test_size=split_2,
            random_state=random_state,
        )

        return (
            features_train,
            target_train,
            features_valid,
            target_valid,
            features_test,
            target_test,
        )

    # Data split into sets without a validation set
    if np.sum([train_share, test_share]) != 1.0:
        raise ValueError("Incorrect sets proportions specified")

    (
        features_train,
        features_test,
        target_train,
        target_test,
    ) = train_test_split(
        features, target, test_size=test_share, random_state=random_state
    )

    return features_train, target_train, features_test, target_test



import functools
import time


def exec_timer(func):
    """Decorator for computing execution time of a function."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        start_time = time.time()
        result = func(*args, **kwargs)
        end_time = time.time()
        execution_time = end_time - start_time
        print(f"Time elasped:{execution_time: .1f} s")
        return result

    return wrapper



from typing import (
    Any,
    Iterable,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from sklearn.base import BaseEstimator
from sklearn.exceptions import NotFittedError
from sklearn.metrics import (
    ConfusionMatrixDisplay,
    classification_report,
    confusion_matrix,
    mean_absolute_error,
    mean_squared_error,
    roc_auc_score,
)
from sklearn.model_selection import RandomizedSearchCV

from .decorators import exec_timer

# Specifying customs types
RandomStateInstance = np.random.mtrand.RandomState


class RandomizedHyperoptRegression(RandomizedSearchCV):
    """
    Wrapper for RandomizedSearchCV with custom defaults
    and additional functionality for running regressions.
    """

    def __init__(
        self,
        estimator: BaseEstimator,
        param_distributions: Union[Sequence[Mapping], Mapping],
        train_dataset: Tuple[Any, Any],
        eval_dataset: Tuple[Any, Any],
        cv: Union[Iterable, int] = 2,
        random_state: Union[RandomStateInstance, int] = 12345,
        scoring: str = "neg_root_mean_squared_error",
        n_iter: int = 5,
        n_jobs: Optional[int] = None,
        return_train_score: bool = True,
    ) -> None:
        """Initializes an instance.

        Args:
            estimator (BaseEstimator): Sklearn model, Gradient Boosting model or
                Pipeline object.
            param_distributions (Union[Sequence[Mapping], Mapping]): Grid of
                hyperparameter names and their values to be varied.
            train_dataset (Tuple[Any, Any]): Tuple of training features and
                training target-vector.
            eval_dataset (Tuple[Any, Any]): Tuple of testing features and
                testing target-vector.
            cv (Union[Iterable, int], optional): Crossvalidator represented
                as an iterator or an integer. Defaults to 2.
            random_state (Union[RandomStateInstance, int], optional): Random seed
                represented as either RandomState instance or an integer.
                Defaults to 12345.
            scoring (str, optional): Scoring metric.
                Defaults to "neg_root_mean_squared_error".
            n_iter (int, optional): Number of hyperparameter combinations
                to consider. Defaults to 5.
            n_jobs (Optional[int], optional): Parallelization of computations.
                Defaults to None.
            return_train_score (bool, optional): Boolean indicating returning
                metric values scores computed on training set. Defaults to True.
        """
        super().__init__(
            estimator=estimator,
            param_distributions=param_distributions,
            cv=cv,
            random_state=random_state,
            scoring=scoring,
            n_iter=n_iter,
            n_jobs=n_jobs,
            return_train_score=return_train_score,
        )
        self.train_dataset = train_dataset
        self.eval_dataset = eval_dataset
        self.train_features, self.train_target = self.train_dataset
        self.eval_features, self.eval_target = self.eval_dataset

    @property
    def _model_name(self):
        """Retrieves last estimator name.

        The function which ends up behaving as a class property
        returns the name of the last estimator in the pipeline or
        the plain model name used.
        """
        try:
            pipeline_steps = self.estimator.named_steps
        except AttributeError:
            model_name = type(self.estimator).__name__
        else:
            pipeline_steps_names = list(pipeline_steps.keys())
            last_estimator_name = pipeline_steps_names[-1]
            last_estimator = self.estimator[last_estimator_name]
            model_name = type(last_estimator).__name__

        return model_name

    @property
    def _relevant_columns(self):
        """Chooses only relevant columns.

        In this case relevant columns in the DataFrame obtained from
        `display_tuning_process()` function involve varied hyperparameters,
        test/train metric values obtained during grid search. Function
        behaves as a class property.
        """
        # Creating the DataFrame from `cv_results_` attribute
        df = pd.DataFrame(self.cv_results_)

        # Choosing specific columns
        columns_to_display = df.columns.str.startswith(("param_", "mean_t"))
        relevant_columns = df.columns[columns_to_display]

        return relevant_columns

    def _fix_col_names(self, df):
        """Fixes columns names in the DataFrame.

        Adjusts column names in the DataFrame returned
        by `display_tuning_process()` function.
        """
        try:
            pipeline_steps = self.estimator.named_steps
        except AttributeError:
            df.columns = df.columns.str.replace("param_", "")
        else:
            pipeline_steps_names = list(pipeline_steps.keys())
            last_estimator_name = pipeline_steps_names[-1]
            col_name = "param_" + last_estimator_name + "__"
            df.columns = df.columns.str.replace(col_name, "")
        finally:
            df = df.round(4)

        return df

    def _verify_est_fitted(self):
        """Checks if the estimator has been fit."""
        if not hasattr(self, "cv_results_"):
            not_fitted_error_msg = (
                f"The {type(self).__name__} instance has not been fitted yet. "
                "Call 'train_crossvalidate' before using this method."
            )
            raise NotFittedError(not_fitted_error_msg)

    @exec_timer
    def train_crossvalidate(self) -> Any:
        """Launches grid search algorithm.

        Using a custom `exec_timer` decorator,
        displaying total grid search time (in
        seconds).
        """

        self.fit(self.train_features, self.train_target)

        model_name = self._model_name

        print(f"Grid search for {model_name} completed.")

    def display_tuning_process(self) -> pd.DataFrame:
        """
        Displaying the results of grid search. Additionally, displays the
        best combination of hyperparameters and scoring metric values.
        """
        # Verifying if the grid search has been launched
        self._verify_est_fitted()

        # Creating the initial DataFrame from `cv_results_` attribute
        cv_results_df = pd.DataFrame(self.cv_results_)

        # Selecting the relevant columns
        relevant_columns = self._relevant_columns
        cv_results_df = cv_results_df[relevant_columns]

        # Fixing column names in the DataFrame
        cv_results_df = self._fix_col_names(cv_results_df)

        # Adjusting the sign in case of regression task
        cv_results_df["mean_test_score"] = -cv_results_df["mean_test_score"]

        if self.return_train_score:
            cv_results_df["mean_train_score"] = -cv_results_df[
                "mean_train_score"
            ]

        return cv_results_df

    @property
    def best_iter(self):
        """Displays the best iteration.

        Based on the DataFrame from `display_tuning_process()`,
        retrieves the best iteration (DataFrame row) and prints
        it out. Behaves as a property.
        """
        # Computing the table with results
        df = self.display_tuning_process()

        # Selecting the best index
        best_iteration = self.best_index_

        # Choosing the best row
        cv_results_best_iter = df.iloc[best_iteration]

        # Retrieving model name (last estimator name)
        model_name = self._model_name

        # Renaming the pd.Series
        cv_results_best_iter = cv_results_best_iter.rename(model_name)

        return cv_results_best_iter

    def compute_regression_stats(
        self, metric: Literal["rmse", "mse", "mae"]
    ) -> float:
        """Computes key regression metrics."""
        # Verifying if the grid search has been launched
        self._verify_est_fitted()
        # Proceed if exception has not been raised
        self.eval_predictions = self.predict(self.eval_features)
        if metric == "rmse":
            rmse = mean_squared_error(
                self.eval_target, self.eval_predictions, squared=False
            )
            return rmse
        elif metric == "mse":
            mse = mean_squared_error(
                self.eval_target, self.eval_predictions, squared=True
            )
            return mse
        elif metric == "mae":
            mae = mean_absolute_error(self.eval_target, self.eval_predictions)
            return mae
        else:
            raise ValueError("Incorrect metric name specified")


class RandomizedHyperoptClassification(RandomizedSearchCV):
    """
    Wrapper for RandomizedSearchCV with custom defaults
    and additional functionality for solving classification
    problems.
    """

    def __init__(
        self,
        estimator: BaseEstimator,
        param_distributions: Union[Sequence[Mapping], Mapping],
        train_dataset: Tuple[Any, Any],
        eval_dataset: Tuple[Any, Any],
        cv: Union[Iterable, int] = 2,
        random_state: Union[RandomStateInstance, int] = 12345,
        scoring: str = "f1",
        n_iter: int = 5,
        n_jobs: Optional[int] = None,
        return_train_score: bool = True,
    ) -> None:
        """Initializes an instance.

        Args:
            estimator (BaseEstimator): Sklearn model, Gradient Boosting model or
                Pipeline object.
            param_distributions (Union[Sequence[Mapping], Mapping]): Grid of
                hyperparameter names and their values to be varied.
            train_dataset (Tuple[Any, Any]): Tuple of training features and
                training target-vector.
            eval_dataset (Tuple[Any, Any]): Tuple of testing features and
                testing target-vector.
            cv (Union[Iterable, int], optional): Crossvalidator represented
                as an iterator or an integer. Defaults to 2.
            random_state (Union[RandomStateInstance, int], optional): Random seed
                represented as either RandomState instance or an integer.
                Defaults to 12345.
            scoring (str, optional): Scoring metric.
                Defaults to "f1".
            n_iter (int, optional): Number of hyperparameter combinations
                to consider. Defaults to 5.
            n_jobs (Optional[int], optional): Parallelization of computations.
                Defaults to None.
            return_train_score (bool, optional): Boolean indicating returning
                metric values scores computed on training set. Defaults to True.
        """
        super().__init__(
            estimator=estimator,
            param_distributions=param_distributions,
            cv=cv,
            random_state=random_state,
            scoring=scoring,
            n_iter=n_iter,
            n_jobs=n_jobs,
            return_train_score=return_train_score,
        )
        self.train_dataset = train_dataset
        self.eval_dataset = eval_dataset
        self.train_features, self.train_target = self.train_dataset
        self.eval_features, self.eval_target = self.eval_dataset

    @property
    def _model_name(self):
        """Retrieves last estimator name.

        The function which ends up behaving as a class property
        returns the name of the last estimator in the pipeline or
        the plain model name used.
        """
        try:
            pipeline_steps = self.estimator.named_steps
        except AttributeError:
            model_name = type(self.estimator).__name__
        else:
            pipeline_steps_names = list(pipeline_steps.keys())
            last_estimator_name = pipeline_steps_names[-1]
            last_estimator = self.estimator[last_estimator_name]
            model_name = type(last_estimator).__name__

        return model_name

    @property
    def _relevant_columns(self):
        """Chooses only relevant columns.

        In this case relevant columns in the DataFrame obtained from
        `display_tuning_process()` function involve varied hyperparameters,
        test/train metric values obtained during grid search. Function
        behaves as a class property.
        """
        df = pd.DataFrame(self.cv_results_)

        # Choosing specific columns
        columns_to_display = df.columns.str.startswith(("param_", "mean_t"))
        relevant_columns = df.columns[columns_to_display]

        return relevant_columns

    def _fix_col_names(self, df):
        """Fixes columns names in the DataFrame.

        Adjusts column names in the DataFrame returned
        by `display_tuning_process()` function.
        """
        try:
            pipeline_steps = self.estimator.named_steps
        except AttributeError:
            df.columns = df.columns.str.replace("param_", "")
        else:
            pipeline_steps_names = list(pipeline_steps.keys())
            last_estimator_name = pipeline_steps_names[-1]
            col_name = "param_" + last_estimator_name + "__"
            df.columns = df.columns.str.replace(col_name, "")
        finally:
            df = df.round(4)

        return df

    def _verify_est_fitted(self):
        """Checks if the estimator has been fit."""
        if not hasattr(self, "cv_results_"):
            not_fitted_error_msg = (
                f"The {type(self).__name__} instance has not been fitted yet. "
                "Call 'train_crossvalidate' before using this method."
            )
            raise NotFittedError(not_fitted_error_msg)

    @exec_timer
    def train_crossvalidate(self) -> Any:
        """Launches grid search algorithm.

        Using a custom `exec_timer` decorator,
        displaying total grid search time (in
        seconds).
        """

        self.fit(self.train_features, self.train_target)

        model_name = self._model_name

        print(f"Grid search for {model_name} completed.")

    def display_tuning_process(self) -> pd.DataFrame:
        """
        Displaying the results of grid search. Additionally, displays the
        best combination of hyperparameters and scoring metric values.
        """
        # Verifying if the grid search has been launched
        self._verify_est_fitted()

        # Creating the initial DataFrame from `cv_results_` attribute
        cv_results_df = pd.DataFrame(self.cv_results_)

        # Selecting the relevant columns
        relevant_columns = self._relevant_columns
        cv_results_df = cv_results_df[relevant_columns]

        # Fixing column names in the DataFrame
        cv_results_df = self._fix_col_names(cv_results_df)

        return cv_results_df

    @property
    def best_iter(self):
        """Displays the best iteration.

        Based on the DataFrame from `display_tuning_process()`,
        retrieves the best iteration (DataFrame row) and prints
        it out. Behaves as a property.
        """
        # Computing the table with results
        df = self.display_tuning_process()

        # Selecting the best index
        best_iteration = self.best_index_

        # Choosing the best row
        cv_results_best_iter = df.iloc[best_iteration]

        # Retrieving model name (last estimator name)
        model_name = self._model_name

        # Renaming the pd.Series
        cv_results_best_iter = cv_results_best_iter.rename(model_name)

        return cv_results_best_iter

    def compute_classification_stats(
        self,
        target_names: Optional[Tuple[str]] = None,
        metric: Optional[str] = None,
    ) -> Optional[float]:
        """Performs computation of classification metrics.

        Args:
            target_names (Tuple[str], optional): Names of classes.
                Defaults to None.
            metric (Optional[str], optional): Name of metric.
                Defaults to None.

        Returns:
            Optional[float]: Floating point number, value of a
            specific classification metric or a classification
            metrics table.
        """
        # Verifying if the grid search has been launched
        self._verify_est_fitted()

        # Computing predictions on the test set
        self.eval_predictions = self.predict(self.eval_features)

        # Return the value of a particular metric
        NoneType = type(None)
        if not isinstance(metric, NoneType):
            # Adding ROC-AUC to metric values to be returned
            if metric == "roc-auc":
                roc_auc_value = roc_auc_score(
                    self.eval_target,
                    self.predict_proba(self.eval_features)[:, 1],
                )

                return roc_auc_value
            # Return some other value of metrics contained in classification_report
            else:
                classification_stats = classification_report(
                    self.eval_target,
                    self.eval_predictions,
                    target_names=target_names,
                    output_dict=True,
                )

                classes_list = list(classification_stats.keys())
                last_class = classes_list[1]
                metric_score = classification_stats[last_class][metric]

                return metric_score
        # Print out a summary of all classification metrics if metric=None
        classification_stats = classification_report(
            self.eval_target, self.eval_predictions, target_names=target_names
        )
        print(classification_stats)

    def plot_confusion_matrix(
        self, label_names: Optional[Tuple[str]] = None
    ) -> Any:
        """Plots a confusion matrix."""
        # Verifying if the grid search has been launched
        self._verify_est_fitted()

        # Computing predictions on the test set
        self.eval_predictions = self.predict(self.eval_features)

        # Plotting confusion matrix
        sns.reset_defaults()
        cm = confusion_matrix(self.eval_target, self.eval_predictions)
        disp = ConfusionMatrixDisplay(
            confusion_matrix=cm, display_labels=label_names
        )
        disp.plot(cmap=plt.cm.Blues)

        # Selecting last estimator name
        model_name = self._model_name

        plt.title(f"Confusion matrix ({model_name})")
        plt.tight_layout()
        plt.show()



from typing import Any, Optional, Sequence, Tuple

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns


def plot_model_comparison(
    results: Sequence[Tuple[str, float]],
    metric_name: str = "metric_name",
    title_name: str = "title_name",
    dot_size: int = 150,
    figure_dims: Tuple[int] = (15, 7),
    xticks_fontsize: int = 15,
    yticks_fontsize: int = 12,
    title_fontsize: int = 20,
    ylabel_fontsize: int = 15,
) -> Any:
    """Provides models performance vizualization.

    Generates a scatterplot with model names and their
    respective metric values for comparison.

    Args:
        results (Sequence[Tuple[str, float]]): Sequence of Tuples with
            model names and metric values.
        metric_name (str, optional): Name of the metric. Defaults to
            "metric_name".
        title_name (str, optional): Title of the plot. Defaults to
            "title_name".
        dot_size (int, optional): Size of scatterplot dots.
            Defaults to 150.
        figure_dims (Tuple[int], optional): Dimensions of the figure.
            Defaults to (15, 7).
        xticks_fontsize (int, optional): Size of xticks on the plot.
            Defaults to 15.
        yticks_fontsize (int, optional): Size of yticks on the plot.
            Defaults to 12.
        title_fontsize (int, optional): Size of the title of the plot.
            Defaults to 20.
        ylabel_fontsize (int, optional): Size of the Y-label.
            Defaults to 15.
    """
    # Separating scores from a sequence of tuples passed
    _, scores = zip(*results)
    # Joining model names with scores
    names_with_scores = ["%s\n%.4f" % (name, loss) for name, loss in results]

    # Making a plot
    plt.figure(figsize=figure_dims)

    plt.scatter(range(len(results)), scores, s=dot_size)

    plt.xticks(
        range(len(results)), names_with_scores, fontsize=xticks_fontsize
    )
    plt.yticks(fontsize=yticks_fontsize)

    plt.title(title_name, fontsize=title_fontsize)
    plt.ylabel(metric_name, fontsize=ylabel_fontsize)

    plt.tight_layout()


def plot_corr_heatmap(
    data: pd.DataFrame,
    target_var: Optional[pd.Series] = None,
    annotation: bool = False,
    annot_num_size: Optional[int] = None,
    heatmap_coloring: Optional[Any] = None,
    upper_triangle: bool = False,
    lower_triangle: bool = False,
) -> Any:
    """Plots a heatmap for the correlation matrix.

    Args:
        data (pd.DataFrame): DataFrame for which a heatmap needs
            to be plotted.
        target_var (Optional[pd.Series], optional): Vector with a
            target variable if needed to include it on the heatmap.
            Defaults to None.
        annotation (bool, optional): Boolean indicator of displaying
            numbers in the heatmap. Defaults to False.
        annot_num_size (Optional[int], optional): Size of the
            figures inside the plot. Defaults to None.
        heatmap_coloring (Optional[Any], optional): Heatmap colormap.
            Defaults to None.
        upper_triangle (bool, optional): Boolean indicator of displaying
            only the upper triangle of the matrix. Defaults to False.
        lower_triangle (bool, optional): Boolean indicator of displaying
            only the lower triangle of the matrix. Defaults to False.
    """
    if target_var is not None:
        data = pd.concat([data, target_var], axis=1)

    corr_matrix = data.corr()

    # Showing upper/lower triangle of a matrix
    if upper_triangle:
        mask = np.zeros_like(corr_matrix)
        mask[np.tril_indices_from(mask)] = True
    elif lower_triangle:
        mask = np.zeros_like(corr_matrix)
        mask[np.triu_indices_from(mask)] = True
    else:
        mask = None

    # Plotting a matrix
    corr_heatmap = sns.heatmap(
        corr_matrix,
        annot=annotation,
        mask=mask,
        annot_kws={"size": annot_num_size},
        cmap=heatmap_coloring,
    )
    corr_heatmap.xaxis.tick_bottom()
    corr_heatmap.yaxis.tick_left()
    corr_heatmap.set(title="Correlation matrix heatmap")
    plt.show()


def plot_class_structure(
    target_var: pd.Series,
    xlabel_name: str = "xlabel_name",
    ylabel_name: str = "ylabel_name",
    title_name: str = "title_name",
) -> Any:
    """Plots the structure of the target variable.

    Args:
        target_var (pd.Series): Target vector.
        xlabel_name (str, optional): Name of the xlabel on the plot.
            Defaults to "xlabel_name".
        ylabel_name (str, optional): Name of the ylabel on the plot.
            Defaults to "ylabel_name".
        title_name (str, optional): Title of the plot.
            Defaults to "title_name".
    """
    target_var.value_counts(normalize=True).plot(
        kind="bar",
        xlabel=xlabel_name,
        ylabel=ylabel_name,
        title=title_name,
    )
    plt.xticks(rotation=0)
    plt.show()
